/*
 * sxid - suid, sgid file and directory checking
 * logging.c - logging functions
 *
 * Copyright (C) 1998 Ben Collins <bcollins@debian.org>
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with sxid; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <strings.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>

#include "structs.h"
#include "globals.h"
#include "notify.h"
#include "logging.h"
#include "search.h"
#include "md5sum.h"

struct file_entry *global_log;
FILE *mail_file;

void
init_file_entries (void)
{
    FILE *log;
    char buf[1024], *s, *tok;

    global_log = NULL;
    log = fopen (config_options.log_file, "r");
    if (log == NULL)
	return;

    while ((fgets (buf, sizeof (buf), log)) != NULL)
    {
	struct file_entry *p;

	buf[strlen (buf) - 1] = '\0';
	s = buf;
	p = (struct file_entry *) malloc (sizeof (struct file_entry));
	p->next = global_log;
	global_log = p;

	tok = strtok (s, " ");
	p->path = strdup (tok);

	p->uid = (uid_t) atoi (strtok (NULL, " "));
	p->gid = (gid_t) atoi (strtok (NULL, " "));
	p->mode = (mode_t) atoi (strtok (NULL, " "));
	p->inode = (ino_t) atoi (strtok (NULL, " "));
	p->old_uid = -1;
	p->old_gid = -1;
	p->old_mode = -1;

	tok = strtok (NULL, " ");
	if (tok != NULL)
	    p->md5 = strdup (tok);
	else
	    p->md5 = NULL;
	p->flags = 0;
	p->flags |= FE_OLD;
    }
    return;
}

void
compare_output (void)
{
    mail_file = tmpfile ();

    if (mail_file == NULL)
    {
	fprintf (stderr, "ERROR: Could not open temp file in compare_output().\n");
	exit (1);
    }
    check_add_rem (mail_file);
    check_attr (mail_file);
    check_u_g (mail_file);
    check_forbidden (mail_file);
    if (config_options.flags & FLAG_LALL)
	do_list_all (mail_file);

    return;
}

struct file_entry *
get_fe_bypath (char *path)
{
    struct file_entry *p;

    p = global_log;

    while (p != NULL)
    {
	if (!strcmp (p->path, path))
	    return p;
	p = p->next;
    }
    p = (struct file_entry *) malloc (sizeof (struct file_entry));
    p->next = global_log;
    p->path = NULL;
    global_log = p;

    return p;
}

int
add_file_entry (char *path, uid_t user_id, gid_t grp_id, mode_t mode,
		ino_t inode, char *md5)
{
    struct file_entry *p;

    p = get_fe_bypath (path);

    if (p->path == NULL)
    {
	p->path = strdup (path);
	p->uid = user_id;
	p->gid = grp_id;
	p->mode = mode;
	p->inode = inode;
	p->old_uid = -1;
	p->old_gid = -1;
	p->old_mode = -1;

	if (md5 != NULL)
	    p->md5 = strdup (md5);
	else
	    p->md5 = NULL;
	p->flags = 0;
	p->flags |= FE_NEW;
    } else
    {
	p->flags = 0;
	if (p->uid != user_id)
	{
	    p->old_uid = p->uid;
	    p->uid = user_id;
	    p->flags |= FE_CHUSER;
	}
	if (p->gid != grp_id)
	{
	    p->old_gid = p->gid;
	    p->gid = grp_id;
	    p->flags |= FE_CHGRP;
	}
	if (p->mode != mode)
	{
	    p->old_mode = p->mode;
	    p->mode = mode;
	    p->flags |= FE_CHMODE;
	}
	if (p->inode != inode)
	{
	    p->inode = inode;
	    p->flags |= FE_CHINODE;
	}
	if (md5 != NULL)
	{
	    if (do_check (path, p->md5))
	    {
		strcpy (p->md5, md5);
		p->flags |= FE_CHSUM;
	    }
	} else
	    p->md5 = NULL;
    }
    return 0;
}

void
save_and_rotate (void)
{
    struct file_entry *p = global_log;
    FILE *new_log;
    int i;

    for (i = config_options.keep_logs - 1; i >= 0; i--)
    {
	char o[1024], n[1024];

	snprintf (o, sizeof (o), "%s%c%d", config_options.log_file, i == 0 ?
		  '\0' : '.', i - 1);
	snprintf (n, sizeof (n), "%s.%d", config_options.log_file, i);

	rename (o, n);
    }

    new_log = fopen (config_options.log_file, "w");
    if (new_log == NULL)
    {
	fprintf (stdout, "ERROR: Could not open new log file %s\n",
		 config_options.log_file);
	exit (1);
    }
    while (p != NULL)
    {
	if (!(p->flags & FE_OLD))
	    fprintf (new_log, "%s %d %d %d %d %s\n", p->path, (int) p->uid,
		     (int) p->gid, (int) p->mode, (int) p->inode,
		     p->md5 != NULL ? p->md5 : "");
	p = p->next;
    }

    fclose (new_log);
    return;
}

char *
fill_space (int count)
{
    char *p;
    /* Always have atleast one space */
    if (count < 1)
	count = 1;
    p = (char *) malloc (count + 1);
    memset (p, ' ', count);
    p[count] = '\0';
    return p;
}

char *
format_ugid (uid_t uid, gid_t gid, mode_t mode)
{
    struct passwd *pw = getpwuid (uid);
    struct group *gr = getgrgid (gid);
    char *ret_val, *temp;

    ret_val = temp = (char *) malloc (32);

    if (mode & S_ISUID)
	temp++[0] = '*';
    if (pw == NULL)
	temp += sprintf (temp, "%d", (int) uid);
    else
	temp += sprintf (temp, "%s", pw->pw_name);

    temp++[0] = '.';

    if (mode & S_ISGID)
	temp++[0] = '*';
    if (gr == NULL)
	temp += sprintf (temp, "%d", (int) gid);
    else
	temp += sprintf (temp, "%s", gr->gr_name);

    return ret_val;
}

char *
super_format_ugid (struct file_entry *p)
{
    struct passwd *pw;
    struct group *gr;
    char *ret_val, *temp;

    ret_val = temp = (char *) malloc (64);

    if (p->flags & FE_CHUSER)
    {
	pw = getpwuid (p->old_uid);
	if (pw == NULL)
	    temp += sprintf (temp, "%d->", (int) p->old_uid);
	else
	    temp += sprintf (temp, "%s->", pw->pw_name);
    }
    pw = getpwuid (p->uid);
    if (pw == NULL)
	temp += sprintf (temp, "%d", (int) p->uid);
    else
	temp += sprintf (temp, "%s", pw->pw_name);

    temp++[0] = '.';

    if (p->flags & FE_CHGRP)
    {
	gr = getgrgid (p->old_gid);
	if (gr == NULL)
	    temp += sprintf (temp, "%d->", (int) p->old_gid);
	else
	    temp += sprintf (temp, "%s->", gr->gr_name);
    }
    gr = getgrgid (p->gid);
    if (gr == NULL)
	temp += sprintf (temp, "%d", (int) p->gid);
    else
	temp += sprintf (temp, "%s", gr->gr_name);

    return ret_val;
}

void
check_add_rem (FILE * mail_file)
{
    struct file_entry *p = global_log;
    char init;

    fprintf (mail_file, "Checking for any additions or removals:\n");

    while (p != NULL)
    {
	init = 0;
	if (p->flags & FE_NEW)
	    init = '+';
	else if (p->flags & FE_OLD)
	    init = '-';
	if (init)
	{
	    char *pmode, *perms;
	    pmode = malloc (8);
	    snprintf (pmode, 8, "%o", p->mode);
	    pmode += strlen (pmode) - 4;
	    perms = format_ugid (p->uid, p->gid, p->mode);
	    if (p->mode & S_IFDIR)
		fprintf (mail_file, " %c %s/%s%s%s%s\n", init, p->path,
			 fill_space (32 - strlen (p->path) - 1), perms,
			 fill_space (20 - strlen (perms)), pmode);
	    else
		fprintf (mail_file, " %c %s%s%s%s%s\n", init, p->path,
			 fill_space (32 - strlen (p->path)), perms,
			 fill_space (20 - strlen (perms)), pmode);

	    changed = 1;
	}
	p = p->next;
    }
    fprintf (mail_file, "\n\n");
    return;
}

void
check_attr (FILE * mail_file)
{
    struct file_entry *p = global_log;
    char init;

    fprintf (mail_file, "Checking for changed attributes or sums/inodes:\n");

    while (p != NULL)
    {
	if (p->flags & FE_CHANGEDATTR)
	{
	    char *pmode, *old_pmode, *perms;

	    if (p->mode & S_IFREG)
		init = p->flags & FE_CHSUM ? 'm' : ' ';
	    else
		init = p->flags & FE_CHINODE ? 'i' : ' ';

	    pmode = malloc (8);

	    snprintf (pmode, 8, "%o", p->mode);
	    pmode += strlen (pmode) - 4;
	    if (p->flags & FE_CHMODE)
	    {
		old_pmode = malloc (10);
		snprintf (old_pmode, 10, "%o->", p->old_mode);
		old_pmode += strlen (old_pmode) - 6;
	    } else
		old_pmode = NULL;

	    perms = super_format_ugid (p);

	    if (p->mode & S_IFDIR)
		fprintf (mail_file, " %c %s/%s%s%s%s%s\n", init, p->path,
			 fill_space (32 - strlen (p->path) - 1), perms,
			 fill_space (20 - strlen (perms)), old_pmode != NULL ?
			 old_pmode : "", pmode);
	    else
		fprintf (mail_file, " %c %s%s%s%s%s%s\n", init, p->path,
			 fill_space (32 - strlen (p->path)), perms,
			 fill_space (20 - strlen (perms)), old_pmode != NULL ?
			 old_pmode : "", pmode);

	    changed = 1;
	}
	p = p->next;
    }
    fprintf (mail_file, "\n\n");
    return;
}

void
check_u_g (FILE * mail_file)
{
    struct file_entry *p = global_log;

    fprintf (mail_file, "Checking for no user/group matches:\n");

    while (p != NULL)
    {
	struct passwd *pw = getpwuid (p->uid);
	struct group *gr = getgrgid (p->gid);
	if ((pw == NULL || gr == NULL) && !(p->flags & FE_OLD))
	{
	    char *pmode, *perms;
	    pmode = malloc (8);
	    snprintf (pmode, 8, "%o", p->mode);
	    pmode += strlen (pmode) - 4;
	    perms = format_ugid (p->uid, p->gid, p->mode);
	    if (p->mode & S_IFDIR)
		fprintf (mail_file, "   %s/%s%s%s%s\n", p->path,
			 fill_space (32 - strlen (p->path) - 1), perms,
			 fill_space (20 - strlen (perms)), pmode);
	    else
		fprintf (mail_file, "   %s%s%s%s%s\n", p->path,
			 fill_space (32 - strlen (p->path)), perms,
			 fill_space (20 - strlen (perms)), pmode);

	    notify_diff = 1;
	}
	p = p->next;
    }
    fprintf (mail_file, "\n\n");
    return;
}

void
check_forbidden (FILE * mail_file)
{
    struct file_entry *p = global_log;
    char init;

    fprintf (mail_file, "Checking for forbidden s[ug]id items:\n");

    while (p != NULL)
    {
	if (forbidden (p->path))
	{
	    char *pmode, *perms;

	    init = ' ';

	    pmode = malloc (8);
	    snprintf (pmode, 8, "%o", p->mode);
	    pmode += strlen (pmode) - 4;

	    perms = format_ugid (p->uid, p->gid, p->mode);

	    if (config_options.flags & FLAG_EFOR)
	    {
		p->mode &= ~(S_ISUID | S_ISGID);
		if (chmod (p->path, p->mode))
		    init = '!';
		else
		{
		    p->flags &= ~FE_NEW;
		    p->flags |= FE_OLD;
		    init = 'r';
		}
	    }
	    if (p->mode & S_IFDIR)
		fprintf (mail_file, " %c %s/%s%s%s%s\n", init, p->path,
			 fill_space (32 - strlen (p->path) - 1), perms,
			 fill_space (20 - strlen (perms)), pmode);
	    else
		fprintf (mail_file, " %c %s%s%s%s%s\n", init, p->path,
			 fill_space (32 - strlen (p->path)), perms,
			 fill_space (20 - strlen (perms)), pmode);

	    notify_diff = 1;
	}
	p = p->next;
    }

    fprintf (mail_file, "\n\n");
    return;
}

void
do_list_all (FILE * mail_file)
{
    struct file_entry *p = global_log;

    fprintf (mail_file, "Complete list of entries:\n");

    while (p != NULL)
    {
	if (!p->flags & FE_OLD)
	{
	    char *pmode, *perms;
	    pmode = malloc (8);
	    snprintf (pmode, 8, "%o", p->mode);
	    pmode += strlen (pmode) - 4;
	    perms = format_ugid (p->uid, p->gid, p->mode);
	    if (p->mode & S_IFDIR)
		fprintf (mail_file, "   %s/%s%s%s%s\n", p->path,
			 fill_space (32 - strlen (p->path) - 1), perms,
			 fill_space (20 - strlen (perms)), pmode);
	    else
		fprintf (mail_file, "   %s%s%s%s%s\n", p->path,
			 fill_space (32 - strlen (p->path)), perms,
			 fill_space (20 - strlen (perms)), pmode);

	}
	p = p->next;
    }
    fprintf (mail_file, "\n\n");
    return;
}
